// File TypeTokenTooks by Joe McFall, July 2007
// Modifications by M. Covington subsequently

using System;
using System.Collections.Generic;
using System.Text;

namespace MATTR   // was TTRLIB
{
    /// <summary>
    /// This class provides tools for finding type token
    /// ratio and moving-average type-token ratio of a list of
    /// tokenized words.
    /// </summary>
    public static class TypeTokenTools
    {
        /// <summary>
        /// Tokenizes a string.
        /// </summary>
        /// <param name="text">A string to tokenize</param>
        /// <returns>A List of tokens</string></returns>
        public static List<string> Tokenize(string text)
        {
            text = text.ToLower();
            //separate out punctuation
            text = text.Replace(".", " . ");
            text = text.Replace(",", " , ");
            text = text.Replace(":", " : ");
            text = text.Replace("(", " ( ");
            text = text.Replace(")", " ) ");
            text = text.Replace("\"", " \" ");
            text = text.Replace("'", " ' ");
            text = text.Replace("[", " [ ");
            text = text.Replace(";", " ; ");
            text = text.Replace("]", " ] ");
            text = text.Replace("{", " } ");
            text = text.Replace("--", " -- ");
            text = text.Replace("-", " - ");   // added by Covington
            text = text.Replace("?", " ? ");
            text = text.Replace("!", " ! ");
            text = text.Replace("$", " $ ");
            text = text.Replace("/", " / ");


            //separate out contractions
            // Changed by Covington so that contractions, except for 's and 'll, go back together as single words.
            // This is to match MontyTagger.
            // This will still differ from CPIDR word counts because CPIDR does not count 's as a word even though it is a token.

            text = text.Replace("' ll", "will");

            text = text.Replace("n ' t", "n't");
            text = text.Replace(" ' re ", "'re ");
            text = text.Replace("i ' m", "i'm");
            text = text.Replace(" ' d", " 'd");
            text = text.Replace(" ' ve", "'ve");

            //and possessives
            //text = text.Replace(" ' s ", " 's ");

            // text = text.ToLower();

            //split at whitespace
            char[] splitter = new char[] { ' ', '\t', '\n', '\r' };
            string[] splitText = text.Split(splitter, StringSplitOptions.RemoveEmptyEntries);
            List<string> tokens = new List<string>(splitText);
            return tokens;
        }


        /// <summary>
        /// Copies a list of strings, removing the elements that do not appear to be words.
        /// Words are assumed to have alphabetic first or second character.
        /// </summary>
        public static List<string> RemoveNonWords(List<string> tokens)
          // Added by M. Covington
        {
          List<string> r = new List<string>(tokens.Count);
          foreach (string s in tokens)
          {
            if ((s.Length >= 1 && Char.IsLetterOrDigit(s[0]))
                 || (s.Length >= 2 && Char.IsLetterOrDigit(s[1]))
               )
            {
              r.Add(s);
            }
          }
          return r;
        }


        /// <summary>
        /// Takes a list of tokens and returns a frequency table
        /// with the count of each token, sorted alphabetically
        /// </summary>
        /// <param name="tokens">An array of string tokens</param>
        /// <returns>A Dictionary representing a word frequency table</returns>
        public static Dictionary<string, int> WordFrequency(List<string> tokens)
        {
            SortedDictionary<string, int> sortedFreqTable = new SortedDictionary<string, int>();
            foreach (string word in tokens)
            {
                if (sortedFreqTable.ContainsKey(word))
                {
                    sortedFreqTable[word]++;
                }
                else
                {
                    sortedFreqTable.Add(word, 1);
                }
            }

            //put all sorted values into a regular dictionary
            Dictionary<string, int> freqTable = new Dictionary<string, int>(sortedFreqTable);

            return freqTable;
        }

        /// <summary>
        /// Takes a list of tokens and returns a frequency table
        /// with the count of each token, sorted alphabetically
        /// </summary>
        /// <param name="tokens">An array of string tokens</param>
        /// <returns>A Dictionary representing a word frequency table</returns>
        public static Dictionary<string, int> WordFrequency(string[] tokens)
          // Same as the foregoing but takes an array as arg.
        {
            List<string> tokensList = new List<string>(tokens);
            return WordFrequency(tokensList);
        }

        /// <summary>
        /// Calculates Type/Token ratio: Number of types divided by number of tokens.
        /// </summary>
        /// <param name="words">A list of tokens</param>
        /// <returns>A double representing the TTR</returns>
        public static double TypeTokenRatio(List<string> words)
        {
            //first get the frequency table
            Dictionary<string, int> freqs = WordFrequency(words);

            //divide the number of words by the number of dictionary entries
            return (double)freqs.Count / words.Count;
        }

        /// <summary>
        /// Calculates Type/Token ratio: Number of types divided by number of tokens.
        /// </summary>
        /// <param name="words">An array of tokens</param>
        /// <returns>A double representing the TTR</returns>
        public static double TypeTokenRatio(string [] words)
        {
            List<string> wordList = new List<string>(words);
            //first get the frequency table
            Dictionary<string, int> freqs = WordFrequency(wordList);

            //divide the number of words by the number of dictionary entries
            return (double) freqs.Count / words.Length;
        }


        /// <summary>
        /// Array in which the TTRs of individual windows are stored for
        /// subsequent use when the MATTR is computed.
        /// Positions near the beginning are left null.
        /// </summary>
        public static double?[] WindowTTRs;


        /// <summary>
        /// Moving average type token ratio.  Computes and returns the MATTR for 
        /// the whole text and also stuffs individual window TTRs into the
        /// array WindowTTRs for possible use later.
        /// </summary>
        /// <param name="words">A list of words</param>
        /// <param name="w">Window size for moving average</param>
        /// <returns>A double representing the moving average TTR</returns>
        public static double MATTR(string text, int w)
        {
            List<string> tokens = Tokenize(text);
            return MATTR(tokens, w);
        }


        /// <summary>
        /// Moving average type token ratio.  Computes and returns the MATTR for 
        /// the whole text and also stuffs individual window TTRs into the
        /// array WindowTTRs for possible use later.
        /// </summary>
        /// <param name="words">A list of words</param>
        /// <param name="w">Window size for moving average</param>
        /// <returns>A double representing the moving average TTR</returns>
        public static double MATTR(List<string> words, int w)
        {

            WindowTTRs = new double?[words.Count];
          
            //if window is greater than count of words, return ttr
            if (w > words.Count)
            {
                return TypeTokenRatio(words);
            }

            Queue<string> window = new Queue<string>();
            //first queue up the first w words in words
            //i will be the indexer through the whole process
            int i = 0;
            for (; i < w; i++)
            {
                window.Enqueue(words[i]);
            }

            //get avg for queue
            double movingAvg = TypeTokenRatio(window.ToArray());

            //get frequency table for queue
            Dictionary<string, int> freqs = WordFrequency(window.ToArray());

            //until you get to the last index
            while (i < words.Count)
            {
                //dequeue the first word
                string firstWord = window.Dequeue();

                //check the freq of this word
                int firstWordFreq = freqs[firstWord];

                //if frequency is 1, remove it from freqs
                if (firstWordFreq == 1)
                {
                    freqs.Remove(firstWord);
                }
                //otherwise, subtract 1 from its frequency
                else
                {
                    freqs[firstWord]--;
                }

                //get the next word fromt the word list
                string nextWord = words[i];

                //see if next word is in the freq table
                //if it is...
                if (freqs.ContainsKey(nextWord))
                {
                    //add one to its count...
                    freqs[nextWord]++;
                }
                else//...
                {
                    //add it to the table
                    freqs.Add(nextWord, 1);
                }

                //enqueue the next word
                window.Enqueue(nextWord);

                //ttr for this window is number of entries in the freq table
                //divided by the window size
                WindowTTRs[i] = (double)freqs.Count / w;

                //add window ttr to moving avg
                movingAvg += (double)WindowTTRs[i];

                //iterate
                i++;
            }

            //avg the moving avg
            //should divide movingAvg by number of
            //iterations (i - w + 1)
            movingAvg /= (i - w + 1);

            return movingAvg;

        }

    }
}
